@@ -19,7 +19,14 @@
 
 #include "superblocks.h"
 
-#define udf_cid_to_enc(cid) ((cid) == 8 ? BLKID_ENC_LATIN1 : (cid) == 16 ? BLKID_ENC_UTF16BE : -1)
+#define is_charset_udf(charspec) ((charspec).type == 0 && strncmp((charspec).info, "OSTA Compressed Unicode", sizeof((charspec).info)) == 0)
+
+#define udf_cid_to_enc(cid) ((cid) == 8 ? UL_ENCODE_LATIN1 : (cid) == 16 ? UL_ENCODE_UTF16BE : -1)
+
+struct charspec {
+	uint8_t	type;
+	char	info[63];
+} __attribute__((packed));
 
 struct dstring128 {
 	uint8_t	cid;
@@ -33,6 +40,12 @@ struct dstring32 {
 	uint8_t	clen;
 } __attribute__((packed));
 
+struct dstring36 {
+	uint8_t	cid;
+	uint8_t	c[34];
+	uint8_t	clen;
+} __attribute__((packed));
+
 struct volume_descriptor {
 	struct descriptor_tag {
 		uint16_t	id;
@@ -62,11 +75,24 @@ struct volume_descriptor {
 			uint32_t	charset_list;
 			uint32_t	max_charset_list;
 			struct dstring128 volset_id;
+			struct charspec	desc_charset;
+			struct charspec	exp_charset;
+			uint32_t	vol_abstract[2];
+			uint32_t	vol_copyright[2];
+			uint8_t		app_id_flags;
+			char		app_id[23];
+			uint8_t		app_id_reserved[8];
+			uint8_t		recording_date[12];
+			uint8_t		imp_id_flags;
+			char		imp_id[23];
+			uint8_t		imp_id_os_class;
+			uint8_t		imp_id_os_id;
+			uint8_t		imp_id_reserved[6];
 		} __attribute__((packed)) primary;
 
 		struct logical_descriptor {
 			uint32_t	seq_num;
-			uint8_t		desc_charset[64];
+			struct charspec	desc_charset;
 			struct dstring128 logvol_id;
 			uint32_t	logical_blocksize;
 			uint8_t		domain_id_flags;
@@ -92,12 +118,28 @@ struct volume_descriptor {
 			uint32_t	num_partitions;
 			uint32_t	imp_use_length;
 		} __attribute__((packed)) logical_vol_integ;
+
+		struct imp_use_volume_descriptor {
+			uint32_t	seq_num;
+			uint8_t 	lvi_id_flags;
+			char		lvi_id[23];
+			uint16_t	lvi_id_udf_rev;
+			uint8_t		lvi_id_os_class;
+			uint8_t		lvi_id_os_id;
+			uint8_t		lvi_id_reserved[4];
+			struct charspec	lvi_charset;
+			struct dstring128 logvol_id;
+			struct dstring36 lvinfo1;
+			struct dstring36 lvinfo2;
+			struct dstring36 lvinfo3;
+		} __attribute__((packed)) imp_use_volume;
 	} __attribute__((packed)) type;
 
 } __attribute__((packed));
 
 #define TAG_ID_PVD  1
 #define TAG_ID_AVDP 2
+#define TAG_ID_IUVD 4
 #define TAG_ID_LVD  6
 #define TAG_ID_TD   8
 #define TAG_ID_LVID 9
@@ -144,7 +186,7 @@ static inline int gen_uuid_from_volset_id(unsigned char uuid[17], struct dstring
 	if (enc == -1)
 		return -1;
 
-	len = blkid_encode_to_utf8(enc, buf, sizeof(buf), volset_id->c, clen);
+	len = ul_encode_to_utf8(enc, buf, sizeof(buf), volset_id->c, clen);
 	if (len < 8)
 		return -1;
 
@@ -182,6 +224,7 @@ static int probe_udf(blkid_probe pr,
 	struct logical_vol_integ_descriptor_imp_use *lvidiu;
 	uint32_t lvid_len = 0;
 	uint32_t lvid_loc = 0;
+	uint64_t s_off;
 	uint32_t bs;
 	uint32_t b;
 	uint16_t type;
@@ -190,12 +233,19 @@ static int probe_udf(blkid_probe pr,
 	size_t i;
 	uint32_t vsd_len;
 	uint16_t udf_rev = 0;
+	int is_udf = 0;
 	int vsd_2048_valid = -1;
 	int have_label = 0;
 	int have_uuid = 0;
 	int have_logvolid = 0;
 	int have_volid = 0;
 	int have_volsetid = 0;
+	int have_applicationid = 0;
+	int have_publisherid = 0;
+
+	/* Session offset */
+	if (blkid_probe_get_hint(pr, "session_offset", &s_off) < 0)
+		s_off = 0;
 
 	/* The block size of a UDF filesystem is that of the underlying
 	 * storage; we check later on for the special case of image files,
@@ -208,16 +258,21 @@ static int probe_udf(blkid_probe pr,
 		if (i != 0 && pbs[0] == pbs[i])
 			continue;
 
+		/* Do not try with block size which is not divisor of session offset */
+		if (s_off % pbs[i])
+			continue;
+
 		/* ECMA-167 2/8.4, 2/9.1: Each VSD is either 2048 bytes long or
 		 * its size is same as blocksize (for blocksize > 2048 bytes)
 		 * plus padded with zeros */
 		vsd_len = pbs[i] > 2048 ? pbs[i] : 2048;
 
-		/* Process 2048 bytes long VSD only once */
-		if (vsd_len == 2048) {
+		/* Process 2048 bytes long VSD on first session only once
+		 * as its location is same for any blocksize */
+		if (s_off == 0 && vsd_len == 2048) {
 			if (vsd_2048_valid == 0)
 				continue;
-			else if (vsd_2048_valid == 1)
+			if (vsd_2048_valid == 1)
 				goto anchor;
 		}
 
@@ -225,7 +280,7 @@ static int probe_udf(blkid_probe pr,
 		for (b = 0; b < 64; b++) {
 			vsd = (struct volume_structure_descriptor *)
 				blkid_probe_get_buffer(pr,
-						UDF_VSD_OFFSET + b * vsd_len,
+						s_off + UDF_VSD_OFFSET + b * vsd_len,
 						sizeof(*vsd));
 			if (!vsd)
 				return errno ? -errno : 1;
@@ -243,40 +298,54 @@ static int probe_udf(blkid_probe pr,
 				 * terminated by the first sector which is not a valid
 				 * descriptor.
 				 * UDF-2.60 2.1.7: UDF 2.00 and lower revisions do not
-				 * have requirement that NSR descritor is in Extended Area
+				 * have requirement that NSR descriptor is in Extended Area
 				 * (between BEA01 and TEA01) and that there is only one
 				 * Extended Area. So do not stop scanning after TEA01. */
 				break;
 		}
 
-		if (vsd_len == 2048)
+		if (s_off == 0 && vsd_len == 2048)
 			vsd_2048_valid = 0;
 
 		/* NSR was not found, try with next block size */
 		continue;
 
 anchor:
-		if (vsd_len == 2048)
+		if (s_off == 0 && vsd_len == 2048)
 			vsd_2048_valid = 1;
 
 		/* Read Anchor Volume Descriptor (AVDP), detect block size */
 		vd = (struct volume_descriptor *)
-			blkid_probe_get_buffer(pr, 256 * pbs[i], sizeof(*vd));
+			blkid_probe_get_buffer(pr, s_off + 256 * pbs[i], sizeof(*vd));
 		if (!vd)
 			return errno ? -errno : 1;
 
 		/* Check that we read correct sector and detected correct block size */
-		if (le32_to_cpu(vd->tag.location) != 256)
-			continue;
+		if (le32_to_cpu(vd->tag.location) == s_off / pbs[i] + 256) {
+			type = le16_to_cpu(vd->tag.id);
+			if (type == TAG_ID_AVDP)
+				goto real_blksz;
+		}
 
-		type = le16_to_cpu(vd->tag.id);
-		if (type == TAG_ID_AVDP)
-			goto real_blksz;
+		/* UDF-2.60: 2.2.3: Unclosed sequential Write-Once media may
+		 * have a single AVDP present at either sector 256 or 512. */
+		vd = (struct volume_descriptor *)
+			blkid_probe_get_buffer(pr, s_off + 512 * pbs[i], sizeof(*vd));
+		if (!vd)
+			return errno ? -errno : 1;
+
+		if (le32_to_cpu(vd->tag.location) == s_off / pbs[i] + 512) {
+			type = le16_to_cpu(vd->tag.id);
+			if (type == TAG_ID_AVDP)
+				goto real_blksz;
+		}
 
 	}
 	return 1;
 
 real_blksz:
+	/* At this stage we detected ISO/IEC 13346 or ECMA-167 filesystem recognition sequence, it does not have to be UDF */
+
 	/* Use the actual block size from here on out */
 	bs = pbs[i];
 
@@ -300,7 +369,7 @@ real_blksz:
 		if (type == TAG_ID_TD)
 			break;
 		if (type == TAG_ID_PVD) {
-			if (!have_volid) {
+			if (!have_volid && is_charset_udf(vd->type.primary.desc_charset)) {
 				int enc = udf_cid_to_enc(vd->type.primary.ident.cid);
 				uint8_t clen = vd->type.primary.ident.clen;
 				if (clen > 0)
@@ -311,7 +380,7 @@ real_blksz:
 					have_volid = !blkid_probe_set_utf8_id_label(pr, "VOLUME_ID",
 							vd->type.primary.ident.c, clen, enc);
 			}
-			if (!have_uuid) {
+			if (!have_uuid && is_charset_udf(vd->type.primary.desc_charset)) {
 				/* VolumeSetIdentifier in UDF 2.01 specification:
 				 * =================================================================================
 				 * 2.2.2.5 dstring VolumeSetIdentifier
@@ -345,7 +414,7 @@ real_blksz:
 				if (gen_uuid_from_volset_id(uuid, &vd->type.primary.volset_id) == 0)
 					have_uuid = !blkid_probe_strncpy_uuid(pr, uuid, sizeof(uuid));
 			}
-			if (!have_volsetid) {
+			if (!have_volsetid && is_charset_udf(vd->type.primary.desc_charset)) {
 				int enc = udf_cid_to_enc(vd->type.primary.volset_id.cid);
 				uint8_t clen = vd->type.primary.volset_id.clen;
 				if (clen > 0)
@@ -356,6 +425,29 @@ real_blksz:
 					have_volsetid = !blkid_probe_set_utf8_id_label(pr, "VOLUME_SET_ID",
 							vd->type.primary.volset_id.c, clen, enc);
 			}
+			if (!have_applicationid) {
+				/* UDF-2.60: 2.2.2.9: This field specifies a valid Entity Identifier identifying the application that last wrote this field */
+				const unsigned char *app_id = (const unsigned char *)vd->type.primary.app_id;
+				size_t app_id_len = strnlen(vd->type.primary.app_id, sizeof(vd->type.primary.app_id));
+				if (app_id_len > 0 && app_id[0] == '*') {
+					app_id++;
+					app_id_len--;
+				}
+				/* When Application Identifier is not set then use Developer ID from Implementation Identifier */
+				if (app_id_len == 0) {
+					/* UDF-2.60: 2.1.5.2: "*Developer ID" refers to an Entity Identifier that uniquely identifies the current implementation */
+					app_id = (const unsigned char *)vd->type.primary.imp_id;
+					app_id_len = strnlen(vd->type.primary.imp_id, sizeof(vd->type.primary.imp_id));
+					if (app_id_len > 0 && app_id[0] == '*') {
+						app_id++;
+						app_id_len--;
+					}
+				}
+				if (app_id_len > 0) {
+					/* UDF-2.60: 2.1.5.2: Values used by UDF for this field are specified in terms of ASCII character strings */
+					have_applicationid = !blkid_probe_set_id_label(pr, "APPLICATION_ID", app_id, app_id_len);
+				}
+			}
 		} else if (type == TAG_ID_LVD) {
 			if (!lvid_len || !lvid_loc) {
 				uint32_t num_partition_maps = le32_to_cpu(vd->type.logical.num_partition_maps);
@@ -365,13 +457,19 @@ real_blksz:
 					lvid_loc = le32_to_cpu(vd->type.logical.lvid_location);
 				}
 			}
-			if (!udf_rev) {
-				/* UDF-2.60: 2.1.5.3: UDF revision field shall indicate revision of UDF document
-				 * We use maximal value from this field and from LVIDIU fields for ID_FS_VERSION */
-				if (strncmp(vd->type.logical.domain_id, "*OSTA UDF Compliant", sizeof(vd->type.logical.domain_id)) == 0)
-					udf_rev = le16_to_cpu(vd->type.logical.udf_rev);
+			if (!is_udf || !udf_rev) {
+				/* UDF-2.60: 2.2.4.3: This field shall indicate that the contents of
+				 * this logical volume conforms to the domain defined in this document.
+				 * This distinguish UDF from all other ISO/IEC 13346 and ECMA-167 filesystems. */
+				if (strncmp(vd->type.logical.domain_id, "*OSTA UDF Compliant", sizeof(vd->type.logical.domain_id)) == 0) {
+					is_udf = 1;
+					/* UDF-2.60: 2.1.5.3: UDF revision field shall indicate revision of UDF document
+					 * We use maximal value from this field and from LVIDIU fields for ID_FS_VERSION */
+					if (!udf_rev)
+						udf_rev = le16_to_cpu(vd->type.logical.udf_rev);
+				}
 			}
-			if (!have_logvolid || !have_label) {
+			if ((!have_logvolid || !have_label) && is_charset_udf(vd->type.logical.desc_charset)) {
 				/* LogicalVolumeIdentifier in UDF 2.01 specification:
 				 * ===============================================================
 				 * 2. Basic Restrictions & Requirements
@@ -413,11 +511,33 @@ real_blksz:
 								vd->type.logical.logvol_id.c, clen, enc);
 				}
 			}
+		} else if (type == TAG_ID_IUVD) {
+			if (!have_publisherid && strncmp(vd->type.imp_use_volume.lvi_id, "*UDF LV Info", sizeof(vd->type.imp_use_volume.lvi_id)) == 0 && is_charset_udf(vd->type.imp_use_volume.lvi_charset)) {
+				/* UDF-2.60: 2.2.7.2.3: Field LVInfo1 could contain information such as Owner Name
+				 * More UDF generating tools set this field to person who creating the filesystem
+				 * therefore its meaning is similar to ISO9660 Publisher Identifier. So for
+				 * compatibility with iso9660 superblock code export this field via PUBLISHER_ID.
+				 */
+				int enc = udf_cid_to_enc(vd->type.imp_use_volume.lvinfo1.cid);
+				uint8_t clen = vd->type.imp_use_volume.lvinfo1.clen;
+				if (clen > 0)
+					--clen;
+				if (clen > sizeof(vd->type.imp_use_volume.lvinfo1.c))
+					clen = sizeof(vd->type.imp_use_volume.lvinfo1.c);
+				if (enc != -1)
+					have_publisherid = !blkid_probe_set_utf8_id_label(pr, "PUBLISHER_ID",
+								vd->type.imp_use_volume.lvinfo1.c, clen, enc);
+			}
 		}
-		if (have_volid && have_uuid && have_volsetid && have_logvolid && have_label && lvid_len && lvid_loc)
+		if (is_udf && have_volid && have_uuid && have_volsetid && have_logvolid && have_label && lvid_len && lvid_loc && have_applicationid && have_publisherid)
 			break;
 	}
 
+	if (!is_udf) {
+		/* We detected some other ISO/IEC 13346 or ECMA-167 filesystem, not UDF */
+		return 1;
+	}
+
 	/* Pick the first logical volume integrity descriptor and read UDF revision */
 	if (lvid_loc && lvid_len >= sizeof(*vd)) {
 		vd = (struct volume_descriptor *)
@@ -463,6 +583,7 @@ real_blksz:
 		 * E.g. number 0x0150 is revision 1.50, number 0x0201 is revision 2.01. */
 		blkid_probe_sprintf_version(pr, "%x.%02x", (unsigned int)(udf_rev >> 8), (unsigned int)(udf_rev & 0xFF));
 
+	blkid_probe_set_fsblocksize(pr, bs);
 	blkid_probe_set_block_size(pr, bs);
 
 	return 0;
@@ -477,13 +598,14 @@ const struct blkid_idinfo udf_idinfo =
 	.flags		= BLKID_IDINFO_TOLERANT,
 	.magics		=
 	{
-		{ .magic = "BEA01", .len = 5, .kboff = 32, .sboff = 1 },
-		{ .magic = "BOOT2", .len = 5, .kboff = 32, .sboff = 1 },
-		{ .magic = "CD001", .len = 5, .kboff = 32, .sboff = 1 },
-		{ .magic = "CDW02", .len = 5, .kboff = 32, .sboff = 1 },
-		{ .magic = "NSR02", .len = 5, .kboff = 32, .sboff = 1 },
-		{ .magic = "NSR03", .len = 5, .kboff = 32, .sboff = 1 },
-		{ .magic = "TEA01", .len = 5, .kboff = 32, .sboff = 1 },
+		/* These magics are generic to all ISO/IEC 13346 and ECMA-167 filesystems, not just UDF */
+		{ .magic = "BEA01", .len = 5, .kboff = 32, .sboff = 1, .hoff = "session_offset" },
+		{ .magic = "BOOT2", .len = 5, .kboff = 32, .sboff = 1, .hoff = "session_offset" },
+		{ .magic = "CD001", .len = 5, .kboff = 32, .sboff = 1, .hoff = "session_offset" },
+		{ .magic = "CDW02", .len = 5, .kboff = 32, .sboff = 1, .hoff = "session_offset" },
+		{ .magic = "NSR02", .len = 5, .kboff = 32, .sboff = 1, .hoff = "session_offset" },
+		{ .magic = "NSR03", .len = 5, .kboff = 32, .sboff = 1, .hoff = "session_offset" },
+		{ .magic = "TEA01", .len = 5, .kboff = 32, .sboff = 1, .hoff = "session_offset" },
 		{ NULL }
 	}
 };
